unit OutputParserUnit;

{*
This source code is provided under the MIT license:
Copyright (C) 2011 by Reinier Olislagers

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*}
{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, DBGrids,
  StdCtrls, EditBtn, fpdataexporter, DB, dbf, fpstdexports, Grids,
  CheckLst, fpspreadsheet {, fpsallformats //presumably for fpspreadsheet use but doesn't seem to be nessary},
  fpcsvexport, fpDBExport {for csv quotestring support},
  fpdbfexport, fptexexport, fpXMLXSDExport;

type
  { TForm1 }

  TForm1 = class(TForm)
    Dbf1: TDbf;
    Exporter: TFPDataExporter;
    ExportFormatChoice: Tlistbox;
    InputFile: TFileNameEdit;
    Datasource1: TDatasource;
    LogOutput: TMemo;
    ImportButton: TButton;
    Label1: TLabel;
    ResultsGrid: TDBGrid;
    StandardExportFormats: TStandardExportFormats;
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure ImportButtonClick(Sender: TObject);
    procedure ResultsGridDrawColumnCell(Sender: TObject; const Rect: TRect;
      DataCol: Integer; Column: TColumn; State: TGridDrawState);
  private
    { private declarations }
    _TempDBF: string; //File where we store our dbf
  public
    { public declarations }
  end;

var
  Form1: TForm1;

implementation

uses ScriptDumpReader;

{$R *.lfm}

{ TForm1 }

procedure TForm1.FormCreate(Sender: TObject);
begin
  // Get a place for our temporary file
  _TempDbf := SysUtils.GetTempFileName + '.dbf';
  //have to specify sysutils to avoid getting the windows one.
  Dbf1.Close;
  // Preselect choice.
  ExportFormatChoice.ItemIndex := 2; //Alpha sort, so should be csv
end;

procedure TForm1.FormClose(Sender: TObject; var CloseAction: TCloseAction);
//<- Clean up
begin
  Dbf1.Close;
  //todo: delete all files that were part of the temp dbf. also dbt etc.
  // there is a function for that somewhere in the dbf code
  //SysUtils.DeleteFile(_TempDbf);
end;

procedure TForm1.ImportButtonClick(Sender: TObject);
//<- Import the data, subsequently immediately export, too.
const
  EXCEL_OUTPUT_FORMAT = sfExcel8; //Excel 97, 2000, XP and 2003;
var
  Reader: TScriptDumpReader;
  SuccesfulImport: boolean;
  CSVExport: TCSVExporter;
  CSVSettings: TCSVFormatSettings;
  DBFExport: TFPDBFExport;
  DBFSettings: TDBFExportFormatSettings;
  ExcelWorkBook: TsWorkBook;
  ExcelWorkSheet: TsWorkSheet;
  ExcelColumn: integer;
  ExcelRow: integer;
  XMLExport: TXMLXSDExporter;
  XMLSettings: TXMLXSDFormatSettings;
  LatexExport: TTeXExporter;
  OutputFileNamePart: string;
begin
  SuccesfulImport := False;
  if (FileExists(InputFile.Text) = False) then
  begin
    ShowMessage('No valid input file. Aborting.');
    exit; //exit function
  end;
  Reader := TScriptDumpReader.Create;
  try
    Datasource1.DataSet := nil;
    LogOutput.Clear;
    Reader.InputFile := InputFile.Text;
    OutputFileNamePart:=Reader.InputFile;
    Reader.DBFFile := _TempDBF;
    Reader.Read; //actual import
    LogOutput.Lines := Reader.ErrorLog;
    SuccesfulImport := True;
  finally
    // Now everything is read in, we can activate the GUI.
    // First release the file, then activate everything
    Reader.Free;
    Datasource1.DataSet := Dbf1;
    Datasource1.Enabled := False;
    ResultsGrid.Enabled := False;
    Dbf1.Close;
    DBf1.FilePathFull := ExtractFilePath(_TempDbf);
    DBf1.TableName := ExtractFileName(_TempDbf);
    Dbf1.ReadOnly := True;
    Dbf1.Open;
    //assert((Dbf1.RecordCount>0), 'there are records in the dataset');
    Datasource1.Enabled := True;
    ResultsGrid.Enabled := True;
    ResultsGrid.Visible := True;
  end;
  if SuccesfulImport = True then
  begin
    case ExportFormatChoice.ItemIndex of
      0: //Access XML
      begin
        XMLExport := TXMLXSDExporter.Create(nil);
        XMLSettings := TXMLXSDFormatSettings.Create(True);
        try
          XMLSettings.DecimalSeparator := char(''); //Don't override decimalseparator;
          XMLSettings.ExportFormat:=AccessCompatible;
          XMLSettings.CreateXSD:=true; //create header info, Access likes this.
          XMLExport.FormatSettings:=XMLSettings;
          XMLExport.Dataset:=Dbf1;
          XMLExport.FileName:=OutputFileNamePart+'.access.xml';
          XMLExport.Execute;
        finally
          XMLSettings.Free;
          XMLExport.Free;
        end;
      end;
      1: //ADO.Net XML
      begin
        XMLExport := TXMLXSDExporter.Create(nil);
        XMLSettings := TXMLXSDFormatSettings.Create(True);
        try
          XMLSettings.ExportFormat:=ADONETCompatible;
          XMLSettings.CreateXSD:=true; //create header info, ADO.Net likes this.
          XMLExport.FormatSettings:=XMLSettings;
          XMLExport.Dataset:=Dbf1;
          XMLExport.FileName:=OutputFileNamePart+'.Net.xml';
          XMLExport.Execute;
        finally
          XMLSettings.Free;
          XMLExport.Free;
        end;
      end;
      2: //CSV
      begin
        CSVExport := TCSVExporter.Create(nil);
        CSVSettings := TCSVFormatSettings.Create(True);
        try
          CSVSettings.FieldDelimiter:=','; //we could use ; also
          CSVSettings.HeaderRow:=true;
          CSVSettings.QuoteStrings:=[qsAlways,qsSpace,qsDelimiter];
          CSVSettings.StringQuoteChar:='"';
          CSVExport.FormatSettings:=CSVSettings;
          CSVExport.Dataset:=Dbf1;
          CSVExport.FileName:=OutputFileNamePart+'.csv';
          CSVExport.Execute;
        finally
          CSVSettings.Free;
          CSVExport.Free;
        end;
      end;
      3: //DBF
      begin
        DBFExport := TFPDBFExport.Create(nil);
        DBFSettings:=TDBFExportFormatSettings.Create(true);
        try
          DBFSettings.AutoRenameFields:=true;
          DBFExport.FormatSettings:=DBFSettings;
          DBFExport.Dataset:=Dbf1;
          DBFExport.FileName:=OutputFileNamePart+'.dbf';
          DBFExport.Execute;
        finally
          DBFSettings.Free;
          DBFExport.Free;
        end;
      end;
      4: //Excel. We use fpSpreadsheet for this.
      { While we could have exported using fpXMLXSDExport,
      the generated XML does not support multiline cells,
      which is essential in this case }
      begin
        ExcelWorkBook:=TsWorkbook.Create;
        ExcelWorkSheet:=ExcelWorkBook.AddWorksheet('Table1');
        try
          ExcelColumn:=0;
          ExcelRow:=0;
          Dbf1.Active:=true;
          { Write out field names on first row }
          for ExcelColumn:=0 to Dbf1.Fields.Count -1 do
            ExcelWorkSheet.WriteUTF8Text(0, ExcelColumn, Dbf1.Fields[ExcelColumn].FieldName);

          { Data on second and further rows }
          ExcelRow:=1; //Start after line of field names
          while not Dbf1.EOF do
          begin
            for ExcelColumn:=0 to Dbf1.Fields.Count -1 do
              ExcelWorkSheet.WriteUTF8Text(ExcelRow, ExcelColumn, Dbf1.Fields[ExcelColumn].AsString);
            Dbf1.Next;
            inc(ExcelRow);
          end;

          // Save the spreadsheet to a file
          ExcelWorkbook.WriteToFile(OutputFileNamePart+STR_EXCEL_EXTENSION, EXCEL_OUTPUT_FORMAT);
        finally
          //ExcelWorkSheet.Free; //Don't do this, the workbook parent will free it.
          ExcelWorkBook.Free;
        end;
      end;
      5: //Latex
      begin
        LatexExport := TTexExporter.Create(nil);
        try
          LatexExport.Dataset:=Dbf1;
          LatexExport.FileName:=OutputFileNamePart+'.tex';
          LatexExport.Execute;
        finally
          LatexExport.Free;
        end;
      end;
      else
        ShowMessage('Unknown export format: ' + IntToStr(ExportFormatChoice.ItemIndex));
    end;
    ShowMessage('Export complete.');
  end
  else
  begin
    ShowMessage('Import was not succesful.');
  end;
end;

procedure TForm1.ResultsGridDrawColumnCell(Sender: TObject; const Rect: TRect;
  DataCol: Integer; Column: TColumn; State: TGridDrawState);
// Draw memo text instead of (Memo)
// Note: grid DefaultDrawing property must be off to avoid duplicate drawing
// To get this, I basically copied the existing DefaultDrawColumnCell procedure
// but tested for memo first. If no memo, pass on to default procedure.
// Maybe slower, more complicated, but it allows for changes in the
// core Lazarus DefaultDrawColumnCell procedure.
var
  OverrideDraw: boolean; //determine if we're going to override drawing
  S: string;
  F: TField;
  DataRow: Integer;
begin
  OverrideDraw:=false;
  try
    F := Column.Field;
    if F.DataType = ftMemo then
    begin
      OverRideDraw:=true;
    end;
  except
    on E: exception do
    begin
      // We might have an inactive datalink or whatever,
      // in that case, pass on our problems to the existing
      // procedure.
      //showMessage('Exception: ' + E.Classname + '/' + E.Message);
      OverRideDraw:=false;
    end;
  end;

  if OverRideDraw=false then
  begin
    // Call normal procedure to handle drawing for us.
    ResultsGrid.DefaultDrawColumnCell(Rect,DataCol,Column,State);
  end
  else
  begin
    // Get to work displaying our memo contents
    // Basically shamelessly ripped from
    // DefaultDrawColumnCell
    // maybe fix something for first/header row
    if F<>nil then
    begin
      //DO display memo ;)
      S := F.DisplayText
    end
    else
    begin
      S := '';
    end;
    //Actual drawing, taken from Grids.DrawCellText coding:
    ResultsGrid.Canvas.TextRect(Rect,Left,Top, S);
  end;
end;

end.

