unit ScriptDumpReader;

{*
This source code is provided under the MIT license:
Copyright (C) 2011 by Reinier Olislagers

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*}

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Dbf;

const
  FieldMarker = '--'; //This text is followed by a field name

const
  RecordEndMarkerPart = 'RecordEnd';
//--RecordEnd marks end of record; partial match so we can deal with it as a field name
const
  OurDBFTableLevel = 7; //Dbase 7. Dbase 4 is not enough as it has 10 character field name

const
  MAXFIELDNAME = 32; //Maximum number of characters permitted in field name.

type
  { TScriptDumpReader }
  TScriptDumpReader = class(TObject)
  const
    RecordCounterName = 'RecordCounter';
  private
    _InputFile: string; // Filename for input file
    _Fields: TStringList; // Fields/columns in input file
    _DBFFile: string;
    _Log: TStringList; //Error/warning log
    _ResultDataset: TDBF; //Storage for results
    function FieldName(Line: string): string;
    function NormalizeFieldName(Name: string): string;
    procedure SetDBFFile(Filename: string);
    procedure ScanForFields;
    procedure ImportData;
  public
    constructor Create;
    destructor Destroy;
      override;
    property InputFile: string read _InputFile write _InputFile;
    property DBFFile: string read _DBFFile write SetDBFFile;
    property ErrorLog: TStringList read _Log;
    procedure Read;
    //imports & stuffs results in supplied TBufDataset
  end; //class

implementation

uses StrUtils {for ansistartstr}, DB {for dataset object}; //internally used units

{ TScriptDumpReader }

procedure TScriptDumpReader.SetDBFFile(Filename: string);
begin
  // Strip all path info. Path is stored in _ResultDataset.
  _DBFFile := Filename;
  _ResultDataset.FilePathFull := ExtractFilePath(_DBFFile);
  _ResultDataset.TableName := ExtractFileName(_DBFFile);
end;

function TScriptDumpReader.FieldName(Line: string): string;
  //<-Returns field name if text line is a field marker; empty string otherwise.
  // Strips out trailing :s
  // Will also detect end of record marker as a field
begin
  if Leftstr(Line, Length(FieldMarker)) = FieldMarker then
  begin
    Line := AnsiMidstr(Line, Length(FieldMarker) + 1, Length(Line));
    //strip off fieldmarker
    if AnsiRightStr(Line, 1) = ':' then
    begin
      Line := AnsiLeftStr(Line, Length(Line) - 1); //strip trailing : if present
    end;
    Result := Line;
  end
  else
  begin
    Result := EmptyStr;
  end;
end;

function TScriptDumpReader.NormalizeFieldName(Name: string): string;
  // <- Formats field name to conform to relevant dbase standard.
begin
  Result := AnsiRightStr(AnsiUpperCase(Name), MAXFIELDNAME);
end;

procedure TScriptDumpReader.ScanForFields;
//<- Reads entire input file and scans for all field/column names
// Required to do first so we can set up our TBufDataset correctly
// We're not using filestream because we want to read lines of text
// Not using TStringlist.ReadFromfile due to possible memory constraints

var
  SourceFile: TextFile;
  TextLine: string;
  //See this for dbf details:
  //http://wiki.lazarus.freepascal.org/Lazarus_Tdbf_Tutorial
begin
  AssignFile(SourceFile, _InputFile);
  try
    reset(SourceFile); //open file for reading
    //Set up dataset:
    _ResultDataset.Close; //Needed for redefinition
    _ResultDataset.Exclusive := True;
    //Exclusive needed for DDL operations, handy for writing anyway.
    //Add a RecordNumber field. This will clash if the
    //text file contains a RecordNumber field, too
    _Fields.Add(RecordCounterName);
    _ResultDataset.Fielddefs.Add(RecordCounterName, ftInteger);
    while not EOF(SourceFile) do
    begin
      readln(SourceFile, TextLine);
      // Pick out all field names and add them to list.
      if (FieldName(TextLine) <> EmptyStr) and
        (FieldName(TextLine) <> RecordEndMarkerPart) then
      begin
        TextLine := FieldName(TextLine); //get field name part
        if (_Fields.IndexOf(TextLine) = -1) then
        begin
          // Field doesn't exist, so add it
          // taking care to respect dbf field naming limits
          TextLine := NormalizeFieldName(TextLine);
          _Fields.Add(TextLine); //maybe later on remove _Fields and just use fielddefs
          _ResultDataset.Fielddefs.Add(TextLine, ftMemo);
          //artificial length limit for strings
        end;
      end;
    end; //eof
    _ResultDataset.CreateTable; //apply field updates
  finally
    {$IFDEF DEBUG}
    _Log.Add('Number of fields: ' + IntToStr(_ResultDataset.Fielddefs.Count));
    {$ENDIF}
    closefile(SourceFile);
  end;
end;

procedure TScriptDumpReader.ImportData;
//<- Once fields are defined, use this procedure to read the actual data
var
  SourceFile: TextFile;
  TextLine: string;
  CurrentField: string;
  RecordCounter: integer;
begin
  AssignFile(SourceFile, _InputFile);
  try
    reset(SourceFile); //open file for reading
    CurrentField := EmptyStr;
    RecordCounter := 0;
    _ResultDataset.Open;
    _ResultDataset.Append; //new record
    while not EOF(SourceFile) do
    begin
      readln(SourceFile, TextLine);
      // Pick out all fields and add them to list.
      if FieldName(TextLine) = RecordEndMarkerPart then
      begin
        CurrentField := EmptyStr; //reset for next record
        RecordCounter := RecordCounter + 1;
        _ResultDataset.FieldByName(RecordCounterName).AsInteger := RecordCounter;
        _ResultDataset.Post; //commit current record
        _ResultDataset.Append; //new record, apparently faster than insert
      end;
      if ((FieldName(TextLine) <> EmptyStr) and
        (FieldName(TextLine) <> RecordEndMarkerPart)) then
      begin
        // Field, not end of record
        // Extract field name part respecting dbf limits
        TextLine := NormalizeFieldName(FieldName(TextLine));
        if _Fields.IndexOf(TextLine) <> -1 then
        begin
          // Save field name for processing data in following lines:
          CurrentField := TextLine;
        end
        else
        begin
          // Field is not present in clientdataset, which shouldn't happen
          raise Exception.Create(
            'Trying to read from a field that is not defined in the program dataset. Contact developer for code correction.');
        end;
      end
      else
      begin
        // Data:
        // if not empty line, add data to relevant column
        TextLine := trimright(TextLine); // Spaces on left side might be significant
        if (TextLine <> EmptyStr) and (CurrentField <> EmptyStr) then
        begin
          //todo: add line counter or something?
          if Length(_ResultDataset.FieldByName(CurrentField).AsString) = 0 then
          begin
            _ResultDataset.FieldByName(CurrentField).AsString := TextLine;
            //field was empty
          end
          else
          begin
            _ResultDataset.FieldByName(CurrentField).AsString :=
              _ResultDataset.FieldByName(CurrentField).AsString + #13#10 + TextLine;
            //append line
          end;
        end;
      end;
    end;
    _ResultDataset.Cancel; //don't post last, open record
  finally
    closefile(SourceFile);
    _ResultDataset.Close; //leave it so others can open it (was opened exclusively)
  end;
end;

procedure TScriptDumpReader.Read;
begin
  if _InputFile = EmptyStr then
  begin
    _Log.Add('No filename specified. Import aborted.');
  end
  else
  begin
    _Log.Add('Import started.');
    ScanForFields;
    ImportData;
    _Log.Add('Import finished.');
  end;
end;

constructor TScriptDumpReader.Create;
begin
  inherited;
  _InputFile := EmptyStr;
  _DBFFile := EmptyStr;
  _Fields := TStringList.Create;
  _Log := TStringList.Create;
  _ResultDataset := TDBF.Create(nil);
  _ResultDataset.Close;
  _ResultDataset.TableLevel := OurDBFTableLevel;
end;

destructor TScriptDumpReader.Destroy;
begin
  _Log.Free;
  _Fields.Free;
  _ResultDataset.Close;
  _ResultDataset.Free;
  inherited Destroy;
end;

end.

